from .iresolver import IResolver
from .iservice_cache import IServiceCache
from .iservice_container import IServiceContainer
from .iservice_entry_collection import IServiceEntryCollection
from .iservice_scope import IServiceScope
from .iservice_stack import IScopeStack
from .service_life import ServiceLife
from .service_entry import ServiceEntry


__all__ = [
    "IResolver",
    "IServiceCache",
    "IServiceContainer",
    "IServiceEntryCollection",
    "IServiceScope",
    "IScopeStack",
    "ServiceLife",
    "ServiceEntry",
]


from abc import ABC, abstractmethod
import time
from types import SimpleNamespace
import pytest
import threading

from pyjudo.disposable import IDisposable

# Mock services for testing
class IServiceA(ABC):
    value: str

class IServiceB(ABC):
    service_a: IServiceA

class IServiceC(ABC):
    service_a: IServiceA
    service_b: IServiceB


class ServiceA(IServiceA):
    def __init__(self, value: str = "A"):
        self.value: str = value


class ServiceB(IServiceB):
    def __init__(self, service_a: IServiceA):
        self.service_a: IServiceA = service_a
        self.value: str = "B"


class ServiceC(IServiceC):
    def __init__(self, service_b: IServiceB, service_a: IServiceA):
        self.service_b: IServiceB = service_b
        self.service_a: IServiceA = service_a
        self.value: str = "C"


class CircularService(IServiceA):
    def __init__(self, service_c: IServiceC):
        self.service_c: IServiceC = service_c


class SoftDisposableService(IServiceA):
    def __init__(self):
        self.value = "A"

    def dispose(self):
        self.value = "disposed"


class HardDisposableService(IServiceA, IDisposable):
    def __init__(self):
        self.value = "A"

    def do_dispose(self):
        self.value = "disposed"


class ISingletonService(ABC):
    @abstractmethod
    def do_something(self) -> str:
        pass


class SingletonService(ISingletonService):
    _instance_count = 0
    _lock = threading.Lock()

    def __init__(self):
        with SingletonService._lock:
            SingletonService._instance_count += 1
        # Simulate some initialization delay
        time.sleep(0.1)

    def do_something(self) -> str:
        return "Singleton Service Operation"
    
    @classmethod
    def get_instance_count(cls) -> int:
        with cls._lock:
            return cls._instance_count

class IScopedService(ABC):
    @abstractmethod
    def do_something(self) -> str:
        pass

class IRepository(ABC):
    @abstractmethod
    def get_data(self) -> str:
        pass

class Repository(IRepository):
    def get_data(self) -> str:
        return "data"

class ScopedService(IScopedService):
    def __init__(self, repository: IRepository):
        self.repository = repository

    def do_something(self) -> str:
        return f"Scoped Service with {self.repository.get_data()}"

class ITransientService(ABC):
    @abstractmethod
    def do_something(self) -> str:
        pass

class TransientService(ITransientService):
    def __init__(self):
        # Simulate some initialization delay
        time.sleep(0.05)

    def do_something(self) -> str:
        return "Transient Service Operation"



@pytest.fixture
def services():
    return SimpleNamespace(
        IServiceA=IServiceA,
        IServiceB=IServiceB,
        IServiceC=IServiceC,
        ISingletonService=ISingletonService,
        IScopedService=IScopedService,
        ITransientService=ITransientService,
        IRepository=IRepository,
        ServiceA=ServiceA,
        ServiceB=ServiceB,
        ServiceC=ServiceC,
        CircularService=CircularService,
        SoftDisposableService=SoftDisposableService,
        HardDisposableService=HardDisposableService,
        SingletonService=SingletonService,
        ScopedService=ScopedService,
        TransientService=TransientService,
        Repository=Repository,
    )


from abc import ABC, abstractmethod

from .service_entry import ServiceEntry


class IServiceEntryCollection(ABC):
    @abstractmethod
    def get[T](self, interface: type[T]) -> ServiceEntry[T]: ...

    @abstractmethod
    def set[T](self, interface: type[T], entry: ServiceEntry[T]) -> None: ...

    @abstractmethod
    def remove[T](self, key: type[T]) -> None: ...

    @abstractmethod
    def __contains__(self, key: type) -> bool: ...


import logging
import threading
from typing import override

from pyjudo.core import IServiceScope, IScopeStack
from pyjudo.exceptions import ServiceScopeError


class ScopeStack(IScopeStack):
    def __init__(self) -> None:
        self.__scopes_stack = threading.local()
        self.__lock = threading.Lock()
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)

    @property
    def _scope_stack(self) -> list[IServiceScope]:
        if not hasattr(self.__scopes_stack, "scopes"):
            self.__scopes_stack.scopes = []
        return self.__scopes_stack.scopes

    @override
    def get_current(self) -> IServiceScope | None:
        with self.__lock:
            if not self._scope_stack:
                return None
            return self._scope_stack[-1]

    @override
    def push(self, scope: IServiceScope) -> None:
        with self.__lock:
            self._scope_stack.append(scope)
            self._logger.debug("Pushed new scope to stack.")

    @override
    def pop(self) -> None:
        with self.__lock:
            try:
                _ = self._scope_stack.pop()
                self._logger.debug("Popped scope from stack.")
            except IndexError:
                raise ServiceScopeError("No scope available to pop.")


from collections.abc import Callable

from .service_life import ServiceLife


class ServiceEntry[T]:
    """
    Represents a service entry in the container.
    """

    constructor: type[T] | Callable[..., T]
    lifetime: ServiceLife

    def __init__(self, constructor: type[T] | Callable[..., T], lifetime: ServiceLife):
        self.constructor = constructor
        self.lifetime = lifetime


from abc import ABC, abstractmethod
from functools import partial
from typing import Any, Callable, Self

from .service_life import ServiceLife
from .iservice_scope import IServiceScope


class IServiceContainer(ABC):
    @abstractmethod
    def register[T](
        self,
        interface: type[T],
        constructor: type[T] | Callable[..., T],
        lifetime: ServiceLife = ServiceLife.TRANSIENT,
    ) -> None: ...

    @abstractmethod
    def unregister(self, interface: type) -> None: ...

    @abstractmethod
    def add_transient[T](
        self, interface: type[T], constructor: type[T] | Callable[..., T]
    ) -> Self: ...

    @abstractmethod
    def add_scoped[T](
        self, interface: type[T], constructor: type[T] | Callable[..., T]
    ) -> Self: ...

    @abstractmethod
    def add_singleton[T](
        self, interface: type[T], constructor: type[T] | Callable[..., T]
    ) -> Self: ...

    @abstractmethod
    def get[T](self, interface: type[T], **overrides: Any) -> T: ...

    @abstractmethod
    def get_factory[T](self, interface: type[T]) -> Callable[..., T]: ...

    @abstractmethod
    def is_registered(self, interface: type) -> bool: ...

    @abstractmethod
    def create_scope(self) -> IServiceScope: ...

    def __getitem__[T](self, key: type[T]) -> Callable[..., T]:
        return partial(self.get, key)


import inspect
import logging
import threading
from typing import Any, Callable, cast, get_args, get_origin, override


from pyjudo.core import IResolver, IServiceEntryCollection, IServiceCache, IScopeStack, ServiceLife
from pyjudo.exceptions import (
    ServiceCircularDependencyError,
    ServiceResolutionError,
    ServiceScopeError,
)
from pyjudo.factory import Factory, FactoryProxy


class Resolver(IResolver):
    def __init__(
        self,
        service_entry_collection: IServiceEntryCollection,
        singleton_cache: IServiceCache,
        scope_stack: IScopeStack,
    ) -> None:
        self.__resolution_stack = threading.local()
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)

        self.service_entry_collection = service_entry_collection
        self.singleton_cache = singleton_cache
        self.scope_stack = scope_stack

    @property
    def _resolution_stack(self) -> set[type]:
        if not hasattr(self.__resolution_stack, "stack"):
            self.__resolution_stack.stack = set()
            self._logger.debug("Initialized a new resolution stack for the thread.")
        return self.__resolution_stack.stack

    @override
    def resolve_anonymous[T](self, constructor: Callable[..., T], overrides: dict[str, Any], binding: Any | None = None) -> T:
        return self._create_instance(constructor, overrides, binding)

    @override
    def resolve[T](self, interface: type[T], overrides: dict[str, Any]) -> T:
        if interface in self._resolution_stack:
            raise ServiceCircularDependencyError(
                f"Circular dependency detected for '{interface.__name__}'"
            )

        _ = self._resolution_stack.add(interface)
        self._logger.debug(f"Resolving service '{interface.__name__}'")

        try:
            entry = self.service_entry_collection.get(interface)

            match entry.lifetime:
                case ServiceLife.SINGLETON:
                    return self._get_singleton(interface, entry.constructor, overrides)
                case ServiceLife.SCOPED:
                    return self._get_scoped(interface, entry.constructor, overrides)
                case ServiceLife.TRANSIENT:
                    return self._get_transient(interface, entry.constructor, overrides)
        finally:
            self._resolution_stack.remove(interface)

    def _get_singleton[T](
        self,
        interface: type[T],
        constructor: type[T] | Callable[..., T],
        overrides: dict[str, Any],
    ) -> T:
        instance = self.singleton_cache.get(interface)

        if instance is None:
            instance = self._create_instance(constructor, overrides)
            self.singleton_cache.add(interface, instance)
        elif overrides: # already exists and overrides are specified
            raise ServiceResolutionError(
                f"Singleton service '{interface.__name__}' already exists. Cannot specify overrides."
            )
        return instance

    def _get_scoped[T](
        self,
        interface: type[T],
        constructor: type[T] | Callable[..., T],
        overrides: dict[str, Any],
    ) -> T:
        scope = self.scope_stack.get_current()

        if scope is None:
            raise ServiceScopeError("No scope available to resolve scoped services.")

        instance = scope.get_instance(interface)

        if instance is None:
            instance = self._create_instance(constructor, overrides)
            scope.add_instance(interface, instance)
        return instance

    def _get_transient[T](
        self,
        _interface: type[T],
        constructor: type[T] | Callable[..., T],
        overrides: dict[str, Any],
    ) -> T:
        return self._create_instance(constructor, overrides)

    def _create_instance[T](
        self, constructor: type[T] | Callable[..., T], overrides: dict[str, Any], binding: Any | None = None
    ) -> T:
        if inspect.isclass(constructor):
            signature = inspect.signature(constructor.__init__)
            parameters = list(signature.parameters.values())[1:] # Skip 'self'
        else:
            signature = inspect.signature(constructor)
            parameters = list(signature.parameters.values())

        constructor_args: list[Any] = []
        constructor_kwargs: dict[str, Any] = {}

        for param in parameters:
            name = param.name

            if name in ("self", "cls") and binding is not None:
                constructor_args.insert(0, binding)
                continue

            if param.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD):
                continue # Skip *args and **kwargs

            origin = get_origin(param.annotation)
            args = get_args(param.annotation)

            if origin is Factory and args:
                interface = args[0]
                resolved = FactoryProxy(self, interface)
            elif name in overrides:
                resolved = overrides[name]
            elif param.annotation in self.service_entry_collection:
                resolved = self.resolve(param.annotation, {})
            elif param.default != inspect.Parameter.empty:
                resolved = param.default
            else:
                raise ServiceResolutionError(
                    f"Unable to resolve dependency '{name}' for '{constructor.__name__}'"
                )
            
            if param.kind == inspect.Parameter.POSITIONAL_ONLY:
                constructor_args.append(resolved)
            elif param.kind in (inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.KEYWORD_ONLY):
                constructor_kwargs[name] = resolved
            else:
                raise Exception("Invalid parameter kind: {param.kind}")
        
        return cast(T, constructor(*constructor_args, **constructor_kwargs))


    def _create_instance__[T](
        self, constructor: type[T] | Callable[..., T], overrides: dict[str, Any], instance: Any | None = None
    ) -> T:
        if inspect.isclass(constructor):
            type_hints = inspect.signature(constructor.__init__).parameters
        else:
            type_hints = inspect.signature(constructor).parameters

        kwargs = {}

        for name, param in type_hints.items():
            if name == "self":
                continue

            # Skip *args and **kwargs
            if param.kind in (
                inspect.Parameter.VAR_POSITIONAL,
                inspect.Parameter.VAR_KEYWORD,
            ):
                continue

            origin = get_origin(param.annotation)
            args = get_args(param.annotation)
            if origin is Factory and args:
                interface = args[0]
                kwargs[name] = FactoryProxy(self, interface)
                continue

            if name in overrides:
                kwargs[name] = overrides[name]
            elif param.annotation in self.service_entry_collection:
                kwargs[name] = self.resolve(param.annotation, {})
            elif param.default != inspect.Parameter.empty:
                kwargs[name] = param.default
            else:
                raise ServiceResolutionError(
                    f"Unable to resolve dependency '{name}' for '{constructor.__name__}'"
                )

        self._logger.debug(f"Creating new instance of '{T}'")

        return cast(T, constructor(**kwargs))  # TODO: Remove cast


import pytest

from pyjudo import ServiceContainer, ServiceLife
from pyjudo.exceptions import ServiceResolutionError


def test_singleton_lifetime(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.ServiceA, ServiceLife.SINGLETON)

    instance1 = container.get(services.IServiceA)
    instance2 = container.get(services.IServiceA)

    assert instance1 is instance2
    assert instance1.value == "A"

def test_overrides_in_singleton(services):
    container = ServiceContainer()
    _ = container.register(services.IServiceA, services.ServiceA, ServiceLife.SINGLETON)

    # First instantiation without overrides
    instance1 = container.get(services.IServiceA)
    assert instance1.value == "A"

    # Singleton should prevent overrides after the instance is created
    with pytest.raises(ServiceResolutionError):
        _ = container.get(services.IServiceA, value="Should Fail")

def test_singleton_thread_safety(services):
    import threading

    container = ServiceContainer()

    _ = container.register(services.ISingletonService, services.SingletonService, ServiceLife.SINGLETON)

    thread_count = 10
    results = []
    barrier = threading.Barrier(thread_count)
    lock = threading.Lock()

    def resolve_singleton():
        # Wait for all threads to be ready
        barrier.wait()
        service = container.get(services.ISingletonService)
        with lock:
            results.append(service)

    threads = [threading.Thread(target=resolve_singleton) for _ in range(thread_count)]

    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()

    # Verify that all resolved services are the same instance
    first_instance = results[0]
    assert all(service is first_instance for service in results[1:]), "Singleton instances are not the same across threads."

    # Verify that only one instance was created
    instance_count = services.SingletonService.get_instance_count()
    assert instance_count == 1, f"SingletonService was instantiated {instance_count} times instead of once."


from typing import override, Any, Protocol

from pyjudo.core import IResolver


class Factory[T](Protocol):
    def __call__(self, **overrides: Any) -> T: ...


class FactoryProxy[T](Factory[T]):
    _resolver: IResolver
    _interface: type[T]

    def __init__(self, resolver: IResolver, interface: type[T]):
        self._resolver = resolver
        self._interface = interface

    @override
    def __call__(self, **overrides: Any) -> T:
        return self._resolver.resolve(self._interface, overrides)

    @override
    def __repr__(self) -> str:
        return f"FactoryProxy({self._interface.__name__})"


from abc import ABC, abstractmethod

from .iservice_scope import IServiceScope


class IScopeStack(ABC):
    @abstractmethod
    def push(self, scope: IServiceScope) -> None: ...

    @abstractmethod
    def pop(self) -> None: ...

    @abstractmethod
    def get_current(self) -> IServiceScope | None: ...


from pyjudo.core import IResolver, IServiceContainer, IScopeStack, IServiceScope, ServiceLife

from .factory import Factory
from .resolver import Resolver
from .scope_stack import ScopeStack
from .service_cache import ServiceCache
from .service_container import ServiceContainer as ServiceContainerBase
from .service_entry_collection import ServiceEntryCollection
from .service_scope import ServiceScope


def ServiceContainer() -> ServiceContainerBase:
    def _scope_factory(stack: IScopeStack, resolver: IResolver) -> IServiceScope:
        cache = ServiceCache()
        return ServiceScope(cache, stack, resolver)

    service_entry_collection = ServiceEntryCollection()
    singleton_cache = ServiceCache()
    scope_stack = ScopeStack()
    resolver = Resolver(
        service_entry_collection=service_entry_collection,
        singleton_cache=singleton_cache,
        scope_stack=scope_stack,
    )
    
    return ServiceContainerBase(
        service_entry_collection=service_entry_collection,
        singleton_cache=singleton_cache,
        scope_stack=scope_stack,
        scope_factory=_scope_factory,
        resolver=resolver,
    )

__all__ = [
    "Factory",
    "ServiceContainer",
    "IServiceContainer",
    "ServiceLife",
]

